/*
** THIS SOFTWARE IS SUBJECT TO COPYRIGHT PROTECTION AND IS OFFERED ONLY
** PURSUANT TO THE 3DFX GLIDE GENERAL PUBLIC LICENSE. THERE IS NO RIGHT
** TO USE THE GLIDE TRADEMARK WITHOUT PRIOR WRITTEN PERMISSION OF 3DFX
** INTERACTIVE, INC. A COPY OF THIS LICENSE MAY BE OBTAINED FROM THE 
** DISTRIBUTOR OR BY CONTACTING 3DFX INTERACTIVE INC(info@3dfx.com). 
** THIS PROGRAM IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER 
** EXPRESSED OR IMPLIED. SEE THE 3DFX GLIDE GENERAL PUBLIC LICENSE FOR A
** FULL TEXT OF THE NON-WARRANTY PROVISIONS.  
** 
** USE, DUPLICATION OR DISCLOSURE BY THE GOVERNMENT IS SUBJECT TO
** RESTRICTIONS AS SET FORTH IN SUBDIVISION (C)(1)(II) OF THE RIGHTS IN
** TECHNICAL DATA AND COMPUTER SOFTWARE CLAUSE AT DFARS 252.227-7013,
** AND/OR IN SIMILAR OR SUCCESSOR CLAUSES IN THE FAR, DOD OR NASA FAR
** SUPPLEMENT. UNPUBLISHED RIGHTS RESERVED UNDER THE COPYRIGHT LAWS OF
** THE UNITED STATES.  
** 
** COPYRIGHT 3DFX INTERACTIVE, INC. 1999, ALL RIGHTS RESERVED
*/

#define MANVERSION              3
#define MANREVISION             10

#ifdef __WIN32__
#define BUILD_NUMBER            0435
#endif

#define BASE_VERSIONSTR "3.10.00.0435"

#ifdef SST1
#   define BASE_HWSTR   "Voodoo Graphics\256"
#   ifdef NT_BUILD
#     define BASE_PRODNAME "Glide\256 for Voodoo Graphics\251 & Windows\256 NT"
#   else
#     define BASE_PRODNAME "Glide\256 for Voodoo Graphics\251 & Windows\256 95/98"
#   endif /* NT_BUILD */
#elif defined(SST96)
#   define BASE_HWSTR   " Voodoo Rush\256"
#   ifdef NT_BUILD
#     define BASE_PRODNAME "Glide\256 for Voodoo Rush\251 & Windows\256 NT"
#   else
#     define BASE_PRODNAME "Glide\256 for Voodoo Rush\251 & Windows\256 95/98"
#   endif /* NT_BUILD */
#elif defined(CVG) || defined(VOODOO2)
#   define BASE_HWSTR   " Voodoo^2\256"
#   ifdef NT_BUILD
#     define BASE_PRODNAME "Glide\256 for Voodoo^2\251 & Windows\256 NT"
#   else
#     define BASE_PRODNAME "Glide\256 for Voodoo^2\251 & Windows\256 95/98"
#   endif /* NT_BUILD */
#elif defined(H3)
#   define BASE_HWSTR   " Voodoo Banshee\256, Voodoo3\256, & Velocity(tm) 100/200"
#     define BASE_PRODNAME "Glide\256 for Voodoo Banshee\256, Voodoo3\256, & Velocity(tm) 100/200; Windows 95 \256, Windows 98 \256, & Windows NT\256"
#else
#   define BASE_HWSTR   "Unknown Chip\256"
#endif

#define VERSIONSTR BASE_VERSIONSTR
#define HWSTR BASE_HWSTR
#define PRODNAME BASE_PRODNAME
